/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    DurbinV2F

Description
    Durbin V2F turbulence model.

SourceFiles
    DurbinV2F.C

\*---------------------------------------------------------------------------*/

#ifndef DurbinV2F_H
#define DurbinV2F_H

#include "RASModel.H"
#include "wallDist.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace incompressible
{
namespace RASModels
{

/*---------------------------------------------------------------------------*\
                           Class DurbinV2F Declaration
\*---------------------------------------------------------------------------*/

class DurbinV2F
:
    public RASModel
{
    // Private data
      // Model Coefficients
        dimensionedScalar Cmu_;
        dimensionedScalar CmuKE_;
        dimensionedScalar Ceps10_;
        dimensionedScalar Ceps11_;
        dimensionedScalar Ceps2_;
        dimensionedScalar C1_;
        dimensionedScalar C2_;
        dimensionedScalar CL_;
        dimensionedScalar CEta_;
        dimensionedScalar oneOnSigmaK_;
        dimensionedScalar oneOnSigmaEps_;
        dimensionedScalar yStarLim_;

        dimensionedScalar f0_;

        wallDist yw_;

      // Fields
        volScalarField k_;
        volScalarField epsilon_;
        volScalarField v2_;
        volScalarField f_;

        volScalarField nut_;


    // Private member functions

        tmp<volScalarField> T() const;
        tmp<volScalarField> L() const;

public:

    //- Runtime type information
    TypeName("DurbinV2F");

    // Constructors

        //- from components
        DurbinV2F
        (
            const volVectorField& U,
            const surfaceScalarField& phi,
            transportModel& transport
        );


    // Destructor

        virtual ~DurbinV2F(){}


    // Member Functions

        virtual tmp<volScalarField> nut() const
        {
            return nut_;
        }

        //- Return the effective diffusivity for k
        tmp<volScalarField> DkEff() const
        {
            return tmp<volScalarField> 
            (
                new volScalarField("DkEff", oneOnSigmaK_*nut_ + nu())
            );
        }

        //- Return the effective diffusivity for epsilon
        tmp<volScalarField> DepsilonEff() const
        {
            return tmp<volScalarField> 
            (
                new volScalarField("DepsilonEff", oneOnSigmaEps_*nut_ + nu())
            );
        }

        virtual tmp<volScalarField> k() const
        {
            return k_;
        }

        //- Note that epsilonTilda is returned as epsilon.
        //  This is the appropriate variable for most purposes.
        virtual tmp<volScalarField> epsilon() const
        {
            return epsilon_;
        }

        virtual tmp<volScalarField> v2() const
        {
            return v2_;
        }

        virtual tmp<volScalarField> f() const
        {
            return f_;
        }

        //-Return the Reynolds stress tensor
        virtual tmp<volSymmTensorField> R() const;
        //-Return the effective stress tensor including the laminar stress
        virtual tmp<volSymmTensorField> devReff() const;
        //-Return the source term for the momentum equation
        virtual tmp<fvVectorMatrix> divDevReff(volVectorField& U) const;

        virtual void correct();

        //- Read turbulenceProperties dictionary
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace RASModels
} // End namespace incompressible 
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
